package de.moritzruth.theaterdsl.value

import kotlinx.collections.immutable.ImmutableSet
import kotlin.math.abs

data class DoubleRangesAndValues(val ranges: ImmutableSet<ClosedFloatingPointRange<Double>>, val values: ImmutableSet<Double>) {
    init {
        require(ranges.isNotEmpty() || values.isNotEmpty()) { "at least one range or value is required" }
    }

    val highest: Double
    val highestBeforeInfinity: Double
    val lowest: Double
    val lowestAfterInfinity: Double

    init {
        var l = Double.POSITIVE_INFINITY
        var lai = Double.POSITIVE_INFINITY
        var h = Double.NEGATIVE_INFINITY
        var hbi = Double.NEGATIVE_INFINITY

        for (v in values + ranges.flatMap { listOf(it.start, it.endInclusive) }) {
            if (v < lai) {
                l = v
                if (v != Double.NEGATIVE_INFINITY) lai = v
            }

            if (v > hbi) {
                h = v
                if (v != Double.POSITIVE_INFINITY) hbi = v
            }
        }

        highest = h
        highestBeforeInfinity = hbi
        lowest = l
        lowestAfterInfinity = lai
    }

    fun getNearestTo(value: Double): Double {
        var nearestValue = 0.0
        var nearestDistance = 0.0
        var isFirst = true

        fun update(v: Double) {
            if (isFirst) {
                nearestValue = v
                nearestDistance = v - value
                isFirst = false
                return
            }

            val distance = abs(v - value)
            if (distance < nearestDistance) {
                nearestValue = v
                nearestDistance = distance
            }
        }

        for (v in values) {
            if (v == value) return v
            update(v)
        }

        for (range in ranges) {
            if (value < range.start) update(range.start)
            else if (value > range.endInclusive) update(range.endInclusive)
            else return value
        }

        return nearestValue
    }

    operator fun contains(value: Double): Boolean = getNearestTo(value) == value
}
